@@ -61,11 +61,11 @@
 (defconstant ARRAY-TOTAL-SIZE-LIMIT most-positive-fixnum)
 
 
-(defconstant CHAR-CODE-LIMIT most-positive-fixnum)
+(defconstant CHAR-CODE-LIMIT 4294967296)
 
 
 
-(defconstant lambda-list-keywords '(&optional &rest &body &key &allow-other-keys))
+(defconstant lambda-list-keywords '(&optional &rest &body &key &allow-other-keys &aux &whole))
 
 
 
@@ -317,6 +317,148 @@
 (defun cddddr (list) (nthcdr 4 list))
 
 
+(defun (setf caar) (newval list)
+  (let ((c (car list)))
+    (rplaca c newval)
+    newval))
+
+(defun (setf cadr) (newval list)
+  (let ((c (cdr list)))
+    (rplaca c newval)
+    newval))
+
+(defun (setf cdar) (newval list)
+  (let ((c (car list)))
+    (rplacd c newval)
+    newval))
+
+(defun (setf cddr) (newval list)
+  (let ((c (cdr list)))
+    (rplacd c newval)
+    newval))
+
+(defun (setf caaar) (newval list)
+  (let ((c (caar list)))
+    (rplaca c newval)
+    newval))
+
+(defun (setf caadr) (newval list)
+  (let ((c (cadr list)))
+    (rplaca c newval)
+    newval))
+
+(defun (setf cadar) (newval list)
+  (let ((c (cdar list)))
+    (rplaca c newval)
+    newval))
+
+(defun (setf caddr) (newval list)
+  (let ((c (cddr list)))
+    (rplaca c newval)
+    newval))
+
+(defun (setf cdaar) (newval list)
+  (let ((c (caar list)))
+    (rplacd c newval)
+    newval))
+
+(defun (setf cdadr) (newval list)
+  (let ((c (cadr list)))
+    (rplacd c newval)
+    newval))
+
+(defun (setf cddar) (newval list)
+  (let ((c (cdar list)))
+    (rplacd c newval)
+    newval))
+
+(defun (setf cdddr) (newval list)
+  (let ((c (cddr list)))
+    (rplacd c newval)
+    newval))
+
+(defun (setf caaaar) (newval list)
+  (let ((c (caaar list)))
+    (rplaca c newval)
+    newval))
+
+(defun (setf caaadr) (newval list)
+  (let ((c (caadr list)))
+    (rplaca c newval)
+    newval))
+
+(defun (setf caadar) (newval list)
+  (let ((c (cadar list)))
+    (rplaca c newval)
+    newval))
+
+(defun (setf caaddr) (newval list)
+  (let ((c (caddr list)))
+    (rplaca c newval)
+    newval))
+
+(defun (setf cadaar) (newval list)
+  (let ((c (cdaar list)))
+    (rplaca c newval)
+    newval))
+
+(defun (setf cadadr) (newval list)
+  (let ((c (cdadr list)))
+    (rplaca c newval)
+    newval))
+
+(defun (setf caddar) (newval list)
+  (let ((c (cddar list)))
+    (rplaca c newval)
+    newval))
+
+(defun (setf cadddr) (newval list)
+  (let ((c (cdddr list)))
+    (rplaca c newval)
+    newval))
+
+(defun (setf cdaaar) (newval list)
+  (let ((c (caaar list)))
+    (rplacd c newval)
+    newval))
+
+(defun (setf cdaadr) (newval list)
+  (let ((c (caadr list)))
+    (rplacd c newval)
+    newval))
+
+(defun (setf cdadar) (newval list)
+  (let ((c (cadar list)))
+    (rplacd c newval)
+    newval))
+
+(defun (setf cdaddr) (newval list)
+  (let ((c (caddr list)))
+    (rplacd c newval)
+    newval))
+
+(defun (setf cddaar) (newval list)
+  (let ((c (cdaar list)))
+    (rplacd c newval)
+    newval))
+
+(defun (setf cddadr) (newval list)
+  (let ((c (cdadr list)))
+    (rplacd c newval)
+    newval))
+
+(defun (setf cdddar) (newval list)
+  (let ((c (cddar list)))
+    (rplacd c newval)
+    newval))
+
+(defun (setf cddddr) (newval list)
+  (let ((c (cdddr list)))
+    (rplacd c newval)
+    newval))
+
+
+
 (defun make-list (len &key initial-element)
   (let (out)
     (dotimes (i len)
@@ -324,6 +466,51 @@
     out))
 
 
+(defun copy-alist (alist)
+  (let (out)
+    (dolist (c alist)
+      (setq out (cons (cons (car c) (cdr c)) out)))
+    (reverse out)))
+
+
+(defun copy-tree (tree)
+  (if (atom tree)
+      tree
+      (cons (car tree) (cdr tree))))
+
+
+(defun tree-equal (t1 t2 &key test test-not)
+  (let ((tst (or test
+		 (if test-not (complement test-not))
+		 #'eql)))
+    (if (and (atom t1) (atom t2))
+	(funcall tst t1 t2)
+	(if (and (consp t1) (consp t2))
+	    (and
+	     (tree-equal (car t1) (car t2) :test test :test-not test-not)
+	     (tree-equal (cdr t1) (cdr t2) :test test :test-not test-not))))))
+
+
+(defun sublis (alist tree &key key test test-not)
+  (let ((tst (or test
+		 (if test-not (complement test-not))
+		 #'eql))
+	(tr (copy-tree tree)))
+    (dolist (c alist)
+      (nsubst (cdr c) (car c) tr :key key :test tst))
+    tr))
+
+
+(defun nsublis (alist tree &key key test test-not)
+  (let ((tst (or test
+		 (if test-not (complement test-not))
+		 #'eql)))
+    (dolist (c alist)
+      (nsubst (cdr c) (car c) tree :key key :test tst))
+    tree))
+
+
+
 (defun endp (l) (null l))
 
 
@@ -427,6 +614,73 @@
 						(progn ,@(cdr clause))))))))))))
 
 
+(defmacro ccase (keyf &rest clauses)
+  (let ((keysym (gensym))
+	outsym)
+    `(let ((,keysym ,keyf))
+       ,(dolist (clause clauses (progn
+				  (setq outsym (append outsym '((t (error 'type-error)))))
+				  (setq outsym (cons 'cond outsym))))
+	  (cond
+	    ((listp (car clause))
+	     (setq outsym (append outsym `(((member ,keysym ',(car clause))
+					    (progn ,@(cdr clause)))))))
+	    (t
+	     (setq outsym (append outsym `(((eql ,keysym ',(car clause))
+					    (progn ,@(cdr clause))))))))))))
+
+
+(defmacro ecase (keyf &rest clauses)
+  (let ((keysym (gensym))
+	outsym)
+    `(let ((,keysym ,keyf))
+       ,(dolist (clause clauses (progn
+				  (setq outsym (append outsym '((t (error 'type-error)))))
+				  (setq outsym (cons 'cond outsym))))
+	  (cond
+		((listp (car clause))
+		 (setq outsym (append outsym `(((member ,keysym ',(car clause))
+						(progn ,@(cdr clause)))))))
+		(t
+		 (setq outsym (append outsym `(((eql ,keysym ',(car clause))
+						(progn ,@(cdr clause))))))))))))
+
+
+(defmacro typecase (keyf &rest clauses)
+  (let ((keysym (gensym))
+	outsym)
+    `(let ((,keysym ,keyf))
+       ,(dolist (clause clauses (setq outsym (cons 'cond outsym)))
+	  (cond ((or (eq (car clause) t)
+		     (eq (car clause) 'otherwise))
+		 (setq outsym (append outsym `((t (progn ,@(cdr clause)))))))
+		(t
+		 (setq outsym (append outsym `(((typep ,keysym ',(car clause))
+						(progn ,@(cdr clause))))))))))))
+
+
+(defmacro ctypecase (keyf &rest clauses)
+  (let ((keysym (gensym))
+	outsym)
+    `(let ((,keysym ,keyf))
+       ,(dolist (clause clauses (progn
+				  (setq outsym (append outsym '((t (error 'type-error)))))
+				  (setq outsym (cons 'cond outsym))))
+	  (setq outsym (append outsym `(((typep ,keysym ',(car clause))
+					 (progn ,@(cdr clause))))))))))
+
+
+(defmacro etypecase (keyf &rest clauses)
+  (let ((keysym (gensym))
+	outsym)
+    `(let ((,keysym ,keyf))
+       ,(dolist (clause clauses (progn
+				  (setq outsym (append outsym '((t (error 'type-error)))))
+				  (setq outsym (cons 'cond outsym))))
+	  (setq outsym (append outsym `(((typep ,keysym ',(car clause))
+					 (progn ,@(cdr clause))))))))))
+
+
 (defmacro return (&optional val)
   `(return-from nil (values-list (multiple-value-list ,val))))
 
@@ -474,88 +728,158 @@
   (not (apply 'every pred sequences)))
 
 
-(defun member (obj list)
-  (let ((subl list))
+(defun member (obj list &key key test test-not)
+  (let ((tst (or test
+	      (if test-not (complement test-not))
+	       #'eql)))
+    (unless key
+      (setq key #'identity))
     (dotimes (i (length list))
-      (if (eql obj (car subl))
-	  (return-from member subl))
-      (setq subl (cdr subl)))))
+      (if (funcall tst obj (funcall key (car list)))
+	  (return-from member list))
+      (setq list (cdr list)))))
 
 
-(defun member-if (pred list)
-  (let ((subl list))
-    (dotimes (i (length list))
-      (if (funcall pred (car subl))
-	  (return-from member-if subl))
-      (setq subl (cdr subl)))))
+(defun member-if (pred list &key key)
+  (unless key
+    (setq key #'identity))
+  (dotimes (i (length list))
+    (if (funcall pred (funcall key (car list)))
+	(return-from member-if list))
+    (setq list (cdr list))))
 
 
-(defun member-if-not (pred list)
-  (member-if (complement pred) list))
+(defun member-if-not (pred list &key key)
+  (member-if (complement pred) list :key key))
 
 
-(defun find (obj seq)
-  (find-if (lambda (x) (eql obj x)) seq))
+(defun find (obj seq &key from-end test test-not (start 0) end key)
+  (let ((tst (or test
+		 (if test-not (complement test-not))
+		 #'eql)))
+    (find-if (lambda (x) (funcall tst obj x)) seq :from-end from-end :start start :end end :key key)))
 
 
-(defun find-if (pred seq)
-  (dotimes (i (length seq) nil)
-    (if (funcall pred (elt seq i))
-	(return-from find-if (elt seq i)))))
+(defun find-if (pred seq &key from-end (start 0) end key)
+  (unless end
+    (setq end (length seq)))
+  (unless key
+    (setq key #'identity))
+  (dotimes (i (- end start))
+    (let ((j (if from-end
+		 (- end i 1)
+		 (+ start i))))
+      (if (funcall pred (funcall key (elt seq j)))
+	  (return-from find-if (elt seq j))))))
 
 
-(defun find-if-not (pred seq)
-  (find-if (complement pred) seq))
+(defun find-if-not (pred seq &key from-end (start 0) end key)
+  (find-if (complement pred) seq :from-end from-end :start start :end end :key key))
 
 
-(defun assoc (obj alist)
-  (assoc-if (lambda (x) (eql obj x)) alist))
+(defun assoc (obj alist &key key test test-not)
+  (let ((tst (or test
+		 (if test-not (complement test-not))
+		 #'eql)))
+    (assoc-if (lambda (x) (funcall tst obj x)) alist :key key)))
 
 
-(defun assoc-if (pred alist)
-  (let ((subl alist))
-    (dotimes (i (length alist))
-      (if (funcall pred (car (car subl)))
-	  (return-from assoc-if (car subl)))
-      (setq subl (cdr subl)))))
+(defun assoc-if (pred alist &key key)
+  (unless key
+    (setq key #'identity))
+  (dotimes (i (length alist))
+    (if (and
+	 (car alist)
+	 (funcall pred (funcall key (caar alist))))
+	(return-from assoc-if (car alist)))
+    (setq alist (cdr alist))))
 
 
-(defun assoc-if-not (pred alist)
-  (assoc-if (complement pred) alist))
+(defun assoc-if-not (pred alist &key key)
+  (assoc-if (complement pred) alist :key key))
 
 
-(defun position (obj seq)
-  (position-if (lambda (x) (eql obj x)) seq))
+(defun rassoc (obj alist &key key test test-not)
+  (let ((tst (or test
+		 (if test-not (complement test-not))
+		 #'eql)))
+    (rassoc-if (lambda (x) (funcall tst obj x)) alist :key key)))
 
 
-(defun position-if (pred seq)
-  (dotimes (i (length seq) nil)
-    (if (funcall pred (elt seq i))
-	(return-from position-if i))))
+(defun rassoc-if (pred alist &key key)
+  (unless key
+    (setq key #'identity))
+  (dotimes (i (length alist))
+    (if (and
+	 (car alist)
+	 (funcall pred (funcall key (cdar alist))))
+	(return-from rassoc-if (car alist)))
+    (setq alist (cdr alist))))
 
 
-(defun position-if-not (pred seq)
-  (position-if (complement pred) seq))
+(defun rassoc-if-not (pred alist &key key)
+  (rassoc-if (complement pred) alist :key key))
 
 
-(defun count (obj seq)
-  (let ((ret 0))
-    (dotimes (i (length seq))
-      (if (eql (elt seq i) obj)
-	  (setq ret (+ 1 ret))))
+(defun position (obj seq &key from-end test test-not (start 0) end key)
+  (let ((tst (or test
+		 (if test-not (complement test-not))
+		 #'eql)))
+    (position-if (lambda (x) (funcall tst obj x)) seq :from-end from-end :start start :end end :key key)))
+
+
+(defun position-if (pred seq &key from-end (start 0) end key)
+  (unless end
+    (setq end (length seq)))
+  (unless key
+    (setq key #'identity))
+  (dotimes (i (- end start))
+    (let ((j (if from-end
+		 (- end i 1)
+		 (+ start i))))
+      (if (funcall pred (elt seq j))
+	  (return-from position-if j)))))
+
+
+(defun position-if-not (pred seq &key from-end (start 0) end key)
+  (position-if (complement pred) seq :from-end from-end :start start :end end :key key))
+
+
+(defun count (obj seq &key from-end (start 0) end key test test-not)
+  (let ((tst (or test
+		 (if test-not (complement test-not))
+		 #'eql))
+	(ret 0))
+    (unless end
+      (setq end (length seq)))
+    (unless key
+      (setq key #'identity))
+    (dotimes (i (- end start))
+      (let ((j (if from-end
+		   (- end i 1)
+		   (+ start i))))
+	(if (funcall tst obj (funcall key (elt seq j)))
+	    (setq ret (+ 1 ret)))))
     ret))
 
 
-(defun count-if (pred seq)
+(defun count-if (pred seq &key from-end (start 0) end key)
+  (unless end
+    (setq end (length seq)))
+  (unless key
+    (setq key #'identity))
   (let ((ret 0))
-    (dotimes (i (length seq))
-      (if (funcall pred (elt seq i))
-	  (setq ret (+ 1 ret))))
+    (dotimes (i (- end start))
+      (let ((j (if from-end
+		   (- end i 1)
+		   (+ start i))))
+	(if (funcall pred (funcall key (elt seq j)))
+	    (setq ret (+ 1 ret)))))
     ret))
 
 
-(defun count-if-not (pred seq)
-  (count-if (complement pred) seq))
+(defun count-if-not (pred seq &key from-end (start 0) end key)
+  (count-if (complement pred) seq :from-end from-end :start start :end end :key key))
 
 
 (defun remove (obj seq)
@@ -578,7 +902,7 @@
   (remove-if (complement pred) seq))
 
 
-(defun remove-duplicates (seq &key from-end test test-not (start 0) end (key #'identity))
+(defun remove-duplicates (seq &key from-end test test-not (start 0) end key)
   (let* ((len (length seq))
 	 (n (if end (- end start) len))
 	 (removed (make-array len))
@@ -590,12 +914,13 @@
 				       'vector
 				       'list)) seq)))
 	 (tst (or test
-		  (if test-not (complement test-not) nil)
-		  #'eql)))
+		  (if test-not (complement test-not))
+		  #'eql))
+	 (k (or key #'identity)))
     (dotimes (i n)
       (dotimes (l (- n i 1))
-	(when (funcall tst (funcall key (elt out (+ start i)))
-		       (funcall key (elt out (+ start i l 1))))
+	(when (funcall tst (funcall k (elt out (+ start i)))
+		       (funcall k (elt out (+ start i l 1))))
 	  (setf (aref removed (+ start i)) t)
 	  (return nil))))
     (setq leftnum (count-if #'not removed))
@@ -618,10 +943,99 @@
   (apply 'remove-duplicates args))
 
 
+(defun substitute (newobj oldobj seq &key from-end test test-not (start 0) end count key)
+  (let ((tst (or test
+		 (if test-not (complement test-not))
+		 #'eql)))
+    (substitute-if newobj (lambda (x) (funcall tst oldobj x)) seq :from-end from-end :start start :end end :count count :key key)))
+
+
+(defun substitute-if (newobj pred seq &key from-end (start 0) end count key)
+  (unless end
+    (setq end (length seq)))
+  (unless key
+    (setq key #'identity))
+  (let ((out (copy-seq seq)))
+    (dotimes (i (- end start))
+      (let ((j (if from-end
+		   (- end i 1)
+		   (+ start i))))
+	(if (and count (= 0 count))
+	    (return-from substitute-if out))
+	(when (funcall pred (funcall key (elt out j)))
+	  (setf (elt out j) newobj)
+	  (if count
+	      (decf count)))))
+    out))
+
+
+(defun substitute-if-not (newobj pred seq &key from-end (start 0) end count key)
+  (substitute-if newobj (complement pred) seq :from-end from-end :start start :end end :count count :key key))
+
+
+(defun nsubstitute (&rest args)
+  (apply 'substitute args))
+
+
+(defun nsubstitute-if (&rest args)
+  (apply 'substitute-if args))
+
+
+(defun nsubstitute-if-not (&rest args)
+  (apply 'substitute-if-not args))
+
+
+
+(defun subst (new old tree &key key test test-not)
+  (nsubst new old (copy-tree tree) :key key :test test :test-not test-not))
+
+
+(defun subst-if (new pred tree &key key)
+  (nsubst-if new pred (copy-tree tree) :key key))
+
+
+(defun subst-if-not (new pred tree &key key)
+  (nsubst-if-not new pred (copy-tree tree) :key key))
+
+
+(defun nsubst (new old tree &key key test test-not)
+  (let ((tst (or test
+		 (if test-not (complement test-not))
+		 #'eql)))
+    (nsubst-if new (lambda (x) (funcall tst old x)) tree :key key)))
+
+
+(defun nsubst-if (new pred tree &key key)
+  (unless key
+    (setq key #'identity))
+  (cond
+    ((funcall pred (funcall key tree))
+     new)
+    ((atom tree)
+     tree)
+    (t
+     (rplaca tree (nsubst-if new pred (car tree)))
+     (rplacd tree (nsubst-if new pred (cdr tree)))
+     tree)))
+
+
+(defun nsubst-if-not (new pred tree &key key)
+  (nsubst-if new (complement pred) tree :key key))
+
+
+
 (defun nreverse (seq)
   (reverse seq))
 
 
+(defun revappend (list tail)
+  (nconc (reverse list) tail))
+
+
+(defun nreconc (list tail)
+  (nconc (reverse list) tail))
+
+
 (defun adjoin (obj list)
   (if (member obj list) list (cons obj list)))
 
@@ -632,6 +1046,16 @@
   seq)
 
 
+(defun replace (seq1 seq2 &key (start1 0) end1 (start2 0) end2)
+  (unless end1
+    (setq end1 (length seq1)))
+  (unless end2
+    (setq end2 (length seq2)))
+  (dotimes (i (min (- end1 start1) (- end2 start2)))
+    (setf (elt seq1 (+ i start1)) (elt seq2 (+ i start2))))
+  seq1)
+
+
 (defmacro push (item place)
   `(setf ,place (cons ,item ,place)))
 
@@ -642,10 +1066,12 @@
      c))
 
 
-(defun set-difference (l1 l2 &key (key #'identity) test test-not)
+(defun set-difference (l1 l2 &key key test test-not)
+  (unless key
+    (setq key #'identity))
   (let ((out (concatenate 'list l1))
 	(tst (or test
-		 (if test-not (complement test-not) nil)
+		 (if test-not (complement test-not))
 		 #'eql)))
     (dolist (l l2)
       (setq out (remove-if (lambda (o) (funcall tst (funcall key o)
@@ -658,6 +1084,133 @@
   (apply 'set-difference args))
 
 
+(defun union (list-1 list-2 &key key test test-not)
+  (remove-duplicates (append list-1 list-2) :key key :test test :test-not test-not))
+
+
+(defun nunion (&rest args)
+  (apply 'union args))
+
+
+(defun intersection (list-1 list-2 &key key test test-not)
+  (unless key
+    (setq key #'identity))
+  (let (out)
+    (dolist (l list-1)
+      (if (member (funcall key l) list-2 :key key :test test :test-not test-not)
+	  (setq out (cons l out))))
+    out))
+
+
+(defun nintersection (&rest args)
+  (apply 'intersection args))
+
+
+(defun set-exclusive-or (list-1 list-2 &key key test test-not)
+  (union (set-difference list-1 list-2 :key key :test test :test-not test-not)
+	 (set-difference list-2 list-1 :key key :test test :test-not test-not)))
+
+
+(defun nset-exclusive-or (&rest args)
+  (apply 'set-exclusive-or args))
+
+
+(defun subsetp (l1 l2 &key key test test-not)
+  (unless key
+    (setq key #'identity))
+  (let ((tst (or test
+		 (if test-not (complement test-not))
+		 #'eql)))
+    (dolist (l l1)
+      (unless (member (funcall key l) l2 :test tst)
+	(return-from subsetp nil)))
+    t))
+
+
+
+(defun mismatch (seq1 seq2 &key from-end test test-not key (start1 0) (start2 0) end1 end2)
+  (unless key
+    (setq key #'identity))
+  (unless end1
+    (setq end1 (length seq1)))
+  (unless end2
+    (setq end2 (length seq2)))
+  (when from-end
+    (setq seq1 (reverse seq1))
+    (let ((tmp end1))
+      (setq end1 (- (length seq1) start1))
+      (setq start1 (- (length seq1) tmp)))
+    (setq seq2 (reverse seq2))
+    (let ((tmp end2))
+      (setq end2 (- (length seq2) start2))
+      (setq start2 (- (length seq2) tmp))))
+  (let ((tst (or test
+		 (if test-not (complement test-not))
+		 #'eql)))
+    (dotimes (i (min (- end1 start1) (- end2 start2)))
+      (unless (funcall tst (funcall key (elt seq1 (+ start1 i)))
+		       (funcall key (elt seq2 (+ start2 i))))
+	(return-from mismatch (if from-end
+				  (- (length seq1) (+ start1 i))
+				  (+ start1 i)))))
+    (if (/= (- end1 start1) (- end2 start2))
+	(if from-end
+	    (- (length seq1) (+ start1 (min (- end1 start1) (- end2 start2))))
+	    (+ start1 (min (- end1 start1) (- end2 start2)))))))
+
+
+
+(defun search (seq1 seq2 &key from-end test test-not key (start1 0) (start2 0) end1 end2)
+  (unless key
+    (setq key #'identity))
+  (unless end1
+    (setq end1 (length seq1)))
+  (unless end2
+    (setq end2 (length seq2)))
+  (if (<= (- end1 start1) 1)
+      (return-from search nil))
+  (when from-end
+    (setq seq1 (reverse seq1))
+    (let ((tmp end1))
+      (setq end1 (- (length seq1) start1))
+      (setq start1 (- (length seq1) tmp)))
+    (setq seq2 (reverse seq2))
+    (let ((tmp end2))
+      (setq end2 (- (length seq2) start2))
+      (setq start2 (- (length seq2) tmp))))
+  (let ((tst (or test
+		 (if test-not (complement test-not))
+		 #'eql)))
+    (dotimes (i (+ 1 (- end2 start2 (length seq1))))
+      (when (funcall tst (funcall key (elt seq1 start1))
+		     (funcall key (elt seq2 (+ start2 i))))
+	(dotimes (j (- end1 start1) (if from-end (return-from search (- (length seq2) i (- end1 start1)))
+					(return-from search i)))
+	  (unless (funcall tst (funcall key (elt seq1 (+ start1 j)))
+			   (funcall key (elt seq2 (+ start2 i j))))
+	    (return nil)))))))
+
+
+
+(defun sort (seq pred &key key)
+  (stable-sort seq pred :key key))
+
+
+(defun stable-sort (seq pred &key key)
+  (unless key
+    (setq key #'identity))
+  (let ((len (length seq)))
+    (dotimes (i len)
+      (dotimes (j (- len i 1))
+	(if (funcall pred (funcall key (elt seq (+ j 1)))
+		     (funcall key (elt seq j)))
+	    (let ((tmp (elt seq j)))
+	      (setf (elt seq j) (elt seq (+ j 1)))
+	      (setf (elt seq (+ j 1)) tmp))))))
+  seq)
+
+
+
 (defun array-rank (array)
   (length (array-dimensions array)))
 
@@ -711,18 +1264,231 @@
   (values nil nil nil))
 
 
+(defun remprop (sym ind)
+  (let ((propl (symbol-plist sym))
+	prevpl)
+    (if (eq (car propl) ind)
+	(progn
+	  (setf (symbol-plist sym) (cddr propl))
+	  (return-from remprop t))
+	(progn
+	  (setq prevpl propl)
+	  (setq propl (cddr propl))
+	  (do nil
+	      ((not propl))
+	    (when (eq (car propl) ind)
+	      (rplacd (cdr prevpl) (cddr propl))
+	      (return-from remprop t))
+	    (setq prevpl propl)
+	    (setq propl (cddr propl)))))))
+
+
+
 (defun char (str ind)
   (aref str ind))
 
+(defun (setf char) (newval str ind)
+  (setf (aref str ind) newval))
 
-(defun string/= (str1 str2)
-  (not (string= str1 str2)))
+(defun schar (str ind)
+  (aref str ind))
+
+(defun (setf schar) (newval str ind)
+  (setf (aref str ind) newval))
+
+(defun bit (arr &rest subs)
+  (apply 'aref arr subs))
+
+(defun (setf bit) (newval arr &rest subs)
+  (apply #'(setf aref) newval arr subs))
+
+(defun sbit (arr &rest subs)
+  (apply 'aref arr subs))
+
+(defun (setf sbit) (newval arr &rest subs)
+  (apply #'(setf aref) newval arr subs))
+
+(defun svref (vec ind)
+  (aref vec ind))
+
+(defun (setf svref) (newval vec ind)
+  (setf (aref vec ind) newval))
+
+
+(defun string= (str1 str2 &key (start1 0) end1 (start2 0) end2)
+  (setq str1 (string str1))
+  (setq str2 (string str2))
+  (unless end1
+    (setq end1 (length str1)))
+  (unless end2
+    (setq end2 (length str2)))
+  (let ((l1 (- end1 start1))
+	(l2 (- end2 start2)))
+    (if (/= l1 l2)
+	(return-from string= nil))
+    (dotimes (i (min l1 l2) t)
+      (unless (char= (elt str1 (+ start1 i)) (elt str2 (+ start2 i)))
+	(return-from string= nil)))))
+
+
+(defun string/= (str1 str2 &key (start1 0) end1 (start2 0) end2)
+  (setq str1 (string str1))
+  (setq str2 (string str2))
+  (unless end1
+    (setq end1 (length str1)))
+  (unless end2
+    (setq end2 (length str2)))
+  (let ((l1 (- end1 start1))
+	(l2 (- end2 start2)))
+    (dotimes (i (min l1 l2))
+      (unless (char= (elt str1 (+ start1 i)) (elt str2 (+ start2 i)))
+	(return-from string/= (+ start1 i))))
+    (if (/= l1 l2)
+	(min l1 l2))))
+
+
+(defun string< (str1 str2 &key (start1 0) end1 (start2 0) end2)
+  (setq str1 (string str1))
+  (setq str2 (string str2))
+  (unless end1
+    (setq end1 (length str1)))
+  (unless end2
+    (setq end2 (length str2)))
+  (let ((l1 (- end1 start1))
+	(l2 (- end2 start2)))
+    (dotimes (i (min l1 l2))
+      (if (char< (elt str1 (+ start1 i)) (elt str2 (+ start2 i)))
+	  (return-from string< (+ start1 i))
+	  (if (char> (elt str1 (+ start1 i)) (elt str2 (+ start2 i)))
+	      (return-from string< nil))))
+    (if (< l1 l2)
+	end1)))
+
+
+(defun string<= (str1 str2 &key (start1 0) end1 (start2 0) end2)
+  (unless end1
+    (setq end1 (length str1)))
+  (let ((ret (funcall 'string< str1 str2 :start1 start1 :end1 end1 :start2 start2 :end2 end2)))
+    (if ret
+	ret
+	(let ((ret (funcall 'string= str1 str2 :start1 start1 :end1 end1 :start2 start2 :end2 end2)))
+	  (if ret
+	      end1)))))
+
+
+(defun string> (str1 str2 &key (start1 0) end1 (start2 0) end2)
+  (setq str1 (string str1))
+  (setq str2 (string str2))
+  (unless end1
+    (setq end1 (length str1)))
+  (unless end2
+    (setq end2 (length str2)))
+  (let ((l1 (- end1 start1))
+	(l2 (- end2 start2)))
+    (dotimes (i (min l1 l2))
+      (if (char> (elt str1 (+ start1 i)) (elt str2 (+ start2 i)))
+	  (return-from string> (+ start1 i))
+	  (if (char< (elt str1 (+ start1 i)) (elt str2 (+ start2 i)))
+	      (return-from string> nil))))
+    (if (> l1 l2)
+	end2)))
+
+
+(defun string>= (str1 str2 &key (start1 0) end1 (start2 0) end2)
+  (unless end1
+    (setq end1 (length str1)))
+  (let ((ret (funcall 'string> str1 str2 :start1 start1 :end1 end1 :start2 start2 :end2 end2)))
+    (if ret
+	ret
+	(let ((ret (funcall 'string= str1 str2 :start1 start1 :end1 end1 :start2 start2 :end2 end2)))
+	  (if ret
+	      end1)))))
+
+
+(defun string-equal (&rest strings)
+  (apply #'string= (mapcar #'string-upcase strings)))
+
+
+(defun string-not-equal (&rest strings)
+  (apply #'string/= (mapcar #'string-upcase strings)))
+
+
+(defun string-lessp (&rest strings)
+  (apply #'string< (mapcar #'string-upcase strings)))
+
+
+(defun string-not-greaterp (&rest strings)
+  (apply #'string<= (mapcar #'string-upcase strings)))
+
+
+(defun string-greaterp (&rest strings)
+  (apply #'string> (mapcar #'string-upcase strings)))
+
+
+(defun string-not-lessp (&rest strings)
+  (apply #'string>= (mapcar #'string-upcase strings)))
+
+
+(defun char/= (&rest chars)
+  (do ((c chars (cdr c)))
+      ((not c) t)
+    (dolist (d (cdr c))
+      (if (char= (car c) d)
+	  (return-from char/= nil)))))
+
+
+(defun char< (&rest chars)
+  (do ((c chars (cdr c)))
+      ((not (cdr c)) t)
+    (if (>= (char-code (car c)) (char-code (cadr c)))
+	(return-from char< nil))))
+
+
+(defun char<= (&rest chars)
+  (do ((c chars (cdr c)))
+      ((not (cdr c)) t)
+    (if (> (char-code (car c)) (char-code (cadr c)))
+	(return-from char<= nil))))
+
+
+(defun char> (&rest chars)
+  (do ((c chars (cdr c)))
+      ((not (cdr c)) t)
+    (if (<= (char-code (car c)) (char-code (cadr c)))
+	(return-from char> nil))))
+
+
+(defun char>= (&rest chars)
+  (do ((c chars (cdr c)))
+      ((not (cdr c)) t)
+    (if (< (char-code (car c)) (char-code (cadr c)))
+	(return-from char>= nil))))
 
 
 (defun char-equal (&rest chars)
   (apply #'char= (mapcar #'char-upcase chars)))
 
 
+(defun char-not-equal (&rest chars)
+  (apply #'char/= (mapcar #'char-upcase chars)))
+
+
+(defun char-lessp (&rest chars)
+  (apply #'char< (mapcar #'char-upcase chars)))
+
+
+(defun char-not-greaterp (&rest chars)
+  (apply #'char<= (mapcar #'char-upcase chars)))
+
+
+(defun char-greaterp (&rest chars)
+  (apply #'char> (mapcar #'char-upcase chars)))
+
+
+(defun char-not-lessp (&rest chars)
+  (apply #'char>= (mapcar #'char-upcase chars)))
+
+
 (defun digit-char (weight &optional (radix 10))
   (if (or (>= weight radix))
       ()
@@ -847,7 +1613,7 @@
 
 (deftype unsigned-byte (&optional s)
   (if (eq s '*)
-      'integer
+      '(integer 0 *)
       `(integer 0 ,(1- (expt 2 s)))))
 
 
@@ -900,6 +1666,9 @@
 (defun vectorp (obj)
   (typep obj 'vector))
 
+(defun simple-vector-p (obj)
+  (typep obj 'simple-vector))
+
 (defun arrayp (obj)
   (typep obj 'array))
 
@@ -909,6 +1678,15 @@
 (defun stringp (obj)
   (typep obj 'string))
 
+(defun simple-string-p (obj)
+  (typep obj 'simple-string))
+
+(defun bit-vector-p (obj)
+  (typep obj 'bit-vector))
+
+(defun simple-bit-vector-p (obj)
+  (typep obj 'simple-bit-vector))
+
 (defun hash-table-p (obj)
   (typep obj 'hash-table))
 
@@ -945,15 +1723,16 @@
   (cond
     ((numberp x) (eql x y))
     ((characterp x) (eql x y))
+    ((and (stringp x) (stringp y))
+     (let ((l (length x)))
+       (if (= l (length y))
+	   (dotimes (i l t)
+	     (if (not (eql (elt x i) (elt y i)))
+		 (return-from equal nil))))))
     ((and (consp x) (consp y))
      (and
-      (let (l)
-	(= (setq l (length x)) (length y))
-	(dotimes (i l)
-	  (if (equal (car x) (car y))
-	      (setq x (cdr x) y (cdr y))
-	      (return-from equal nil)))
-	(equal x y))))
+      (equal (car x) (car y))
+      (equal (cdr x) (cdr y))))
     (t (eq x y))))
 
 
@@ -963,13 +1742,8 @@
     ((and (characterp x) (characterp y)) (char-equal x y))
     ((and (consp x) (consp y))
      (and
-      (let (l)
-	(= (setq l (length x)) (length y))
-	(dotimes (i l)
-	  (if (equalp (car x) (car y))
-	      (setq x (cdr x) y (cdr y))
-	      (return-from equalp nil)))
-	(equalp x y))))
+      (equalp (car x) (car y))
+      (equalp (cdr x) (cdr y))))
     ((and (arrayp x) (arrayp y)) (let ((dimsx (array-dimensions x))
 				       (dimsy (array-dimensions y)))
 				   (and
@@ -996,11 +1770,131 @@
 
 
 (defun mapc (fun &rest lists)
-  (apply #'mapcar (cons fun lists))
+  (apply #'mapcar fun lists)
   (car lists))
 
 
 
+(defun mapcan (fun &rest lists)
+  (apply #'nconc (apply 'mapcar fun lists)))
+
+
+
+(defun maplist (fun &rest lists)
+  (let (out)
+    (do nil
+	(nil)
+      (if (member nil lists)
+	  (return nil))
+      (setq out (cons (apply fun lists) out))
+      (setq lists (mapcar #'cdr lists)))
+    (reverse out)))
+
+
+
+(defun mapl (fun &rest lists)
+  (apply #'maplist fun lists)
+  (car lists))
+
+
+
+(defun mapcon (fun &rest lists)
+  (apply #'nconc (apply 'maplist fun lists)))
+
+
+
+(defun map-into (res fun &rest seqs)
+  (let ((len (apply 'min (mapcar #'length (cons res seqs)))))
+    (dotimes (i len)
+      (setf (elt res i) (apply fun (mapcar (lambda (s) (elt s i)) seqs))))
+    res))
+
+
+(defun reduce (fun seq &key key from-end (start 0) end (initial-value nil initvalprov))
+  (unless end
+    (setq end (length seq)))
+  (unless key
+    (setq key #'identity))
+  (cond
+    ((= (- end start) 0)
+     (if initvalprov
+	 initial-value
+	 (funcall fun)))
+    ((and (= (- end start) 1)
+	  (not initvalprov))
+     (funcall key (elt seq start)))
+    (t
+     (let (out)
+       (dotimes (i (- end start))
+	 (let ((j (if from-end
+		      (- end i 1)
+		      (+ start i))))
+	   (if (= i 0)
+	       (if initvalprov
+		   (setq out (funcall fun initial-value (funcall key (elt seq j))))
+		   (progn
+		     (setq out (funcall key (elt seq j)))))
+	       (if from-end
+		   (setq out (funcall fun (funcall key (elt seq j)) out))
+		   (setq out (funcall fun out (funcall key (elt seq j))))))))
+       out))))
+
+
+(defun merge (restype seq1 seq2 pred &key key)
+  (unless key
+    (setq key #'identity))
+  (let* ((outsz (+ (length seq1) (length seq2)))
+	 (out (cond
+		((subtypep restype 'list)
+		 (make-list outsz))
+		((subtypep restype 'string)
+		 (make-string outsz))
+		(t
+		 (make-array outsz))))
+	 (j 0)
+	 (k 0))
+    (dotimes (i outsz)
+      (if (and
+	   (< j (length seq1))
+	   (or
+	    (>= k (length seq2))
+	    (funcall pred (funcall key (elt seq1 j))
+		     (funcall key (elt seq2 k)))
+	    (and (not (funcall pred (funcall key (elt seq1 j))
+			       (funcall key (elt seq2 k))))
+		 (not (funcall pred (funcall key (elt seq2 k))
+			       (funcall key (elt seq1 j)))))))
+	  (progn
+	    (setf (elt out i) (elt seq1 j))
+	    (incf j))
+	  (progn
+	    (setf (elt out i) (elt seq2 k))
+	    (incf k))))
+    out))
+
+
+
+(defun pathname-host (fn &key case))
+
+(defun pathname-device (fn &key case))
+
+(defun pathname-directory (fn &key case))
+
+(defun pathname-type (fn &key case))
+
+(defun pathname-version (fn &key case))
+
+
+(defun file-author (fn))
+
+
+
+(defmacro with-open-file ((str fn &rest opts) &body forms)
+  `(let ((,str (open ,fn ,@opts)))
+     ,@forms))
+
+
+
 (defun terpri (&optional (out *standard-output*))
   (write-char #\newline out)
   nil)
@@ -1516,6 +2410,11 @@
 
 
 
+(defun abort (&optional cond)
+  (invoke-restart 'abort))
+
+
+
 
 (export '(*features* machine-instance machine-type machine-version
 	  short-site-name long-site-name *modules* provide require describe
@@ -1531,25 +2430,39 @@
 	  first second third fourth fifth sixth seventh eighth ninth tenth rest
 	  caar cadr cdar cddr caaar caadr cadar caddr cdaar cdadr cddar cdddr
 	  caaaar caaadr caadar caaddr cadaar cadadr caddar cadddr cdaaar cdaadr
-	  cdadar cdaddr cddaar cddadr cdddar cddddr make-list endp butlast
-	  nbutlast acons pairlis when unless incf decf and or cond otherwise
-	  case return multiple-value-bind prog prog* every some notany notevery
-	  member member-if member-if-not find find-if find-if-not assoc assoc-if
-	  assoc-if-not position position-if position-if-not count count-if
-	  count-if-not remove remove-if-not delete delete-if delete-if-not
-	  remove-duplicates delete-duplicates nreverse adjoin fill push pop
-	  set-difference nset-difference array-rank array-dimension
+	  cdadar cdaddr cddaar cddadr cdddar cddddr make-list copy-alist
+	  copy-tree tree-equal sublis nsublis endp butlast nbutlast acons
+	  pairlis when unless incf decf and or cond otherwise case ccase ecase
+	  typecase ctypecase etypecase return multiple-value-bind prog prog*
+	  every some notany notevery member member-if member-if-not find find-if
+	  find-if-not assoc assoc-if assoc-if-not rassoc rassoc-if rassoc-if-not
+	  position position-if position-if-not count count-if count-if-not
+	  remove remove-if-not delete delete-if delete-if-not remove-duplicates
+	  delete-duplicates substitute substitute-if substitute-if-not
+	  nsubstitute nsubstitute-if nsubstitute-if-not subst subst-if
+	  subst-if-not nsubst nsubst-if nsubst-if-not nreverse revappend nreconc
+	  adjoin fill replace push pop set-difference nset-difference union
+	  nunion intersection nintersection set-exclusive-or nset-exclusive-or
+	  subsetp mismatch search sort stable-sort array-rank array-dimension
 	  array-total-size array-in-bounds-p upgraded-array-element-type
-	  adjustable-array-p get get-properties char string/= char-equal
-	  digit-char digit-char-p char-int string-upcase string-downcase
-	  string-capitalize nstring-upcase nstring-downcase nstring-capitalize
-	  string-left-trim string-right-trim string-trim defpackage signed-byte
-	  unsigned-byte consp listp symbolp keywordp functionp packagep integerp
-	  rationalp floatp complexp random-state-p characterp standard-char-p
-	  vectorp arrayp sequencep stringp hash-table-p pathnamep streamp realp
-	  numberp check-type macroexpand equal equalp fdefinition complement
-	  mapc terpri write-line write-sequence prin1 princ print do-all-symbols
-	  loop format encode-universal-time))
+	  adjustable-array-p get get-properties remprop char schar bit sbit
+	  svref string= string/= string< string<= string> string>= string-equal
+	  string-not-equal string-lessp string-not-greaterp string-greaterp
+	  string-not-lessp char/= char< char<= char> char>= char-equal
+	  char-not-equal char-lessp char-not-greaterp char-greaterp
+	  char-not-lessp digit-char digit-char-p char-int string-upcase
+	  string-downcase string-capitalize nstring-upcase nstring-downcase
+	  nstring-capitalize string-left-trim string-right-trim string-trim
+	  defpackage signed-byte unsigned-byte consp listp symbolp keywordp
+	  functionp packagep integerp rationalp floatp complexp random-state-p
+	  characterp standard-char-p vectorp simple-vector-p arrayp sequencep
+	  stringp simple-string-p bit-vector-p simple-bit-vector-p hash-table-p
+	  pathnamep streamp realp numberp check-type macroexpand equal equalp
+	  fdefinition complement mapc mapcan maplist mapl mapcon map-into reduce
+	  merge pathname-host pathname-device pathname-directory pathname-type
+	  pathname-version file-author with-open-file terpri write-line
+	  write-sequence prin1 princ print do-all-symbols loop format
+	  encode-universal-time abort))
 
 
 
